﻿// Note: Make sure to set Visual Studio's working directory 
//       (under Project->MazeMaker Properties->Debug) to an absolute path for "WorkingDir",
//       otherwise the program won't be able to load the its data files. (Unfortunately,
//       Visual Studio won't play nice with a relative path, which would be needed to
//       store the setting in version controlled project files.)

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using SemiTwist.Utilities;

namespace MazeMaker
{
    public partial class Form1 : Form
    {
        private Grid grid;
        private Dictionary<string, CellType> cellTypes;
        private Point currCell = new Point();
        private bool isMouseOverGridPanel = false;

        private Size _statusPanelGridSize = new Size();
        private Size StatusPanelGridSize
        {
            get
            {
                return _statusPanelGridSize;
            }
            set
            {
                _statusPanelGridSize = value;
                UpdateStatusLabel(grid);
            }
        }

        #region Initilization Data for cellTypes (not to be used for anything else)
        private CellType[] cellTypeInitData =
        {
            new CellType("block_empty.bmp", "Empty", 1),
            new CellType("block_wall.bmp", "Wall", 2),
        };
        #endregion

        public Form1()
        {
            InitializeComponent();
            SetupCellTypes();

            grid = new Grid(gridPanel, cellTypes["Empty"]);
            grid.preResize += new Grid.ResizeEventHandler(grid_preResize);
            grid.onResizeOutOfMemory += new Grid.ExceptionEventHandler(grid_onResizeOutOfMemory);
            grid.autoDrawOnSetCell = false;
            grid.Resize(32, 32);
            UpdateStatusLabel(grid);

            gridPanel.MouseMove += new MouseEventHandler(gridPanel_MouseMove);
            gridPanel.MouseLeave += new EventHandler(gridPanel_MouseLeave);
        }

        private void grid_preResize(object sender, Size newSize)
        {
            StatusPanelGridSize = newSize;
            WidthSet.Text = newSize.Width.ToString();
            HeightSet.Text = newSize.Height.ToString();
        }

        void grid_onResizeOutOfMemory(object sender, Exception e)
        {
            Size attemptedSize = StatusPanelGridSize;
            StatusPanelGridSize = grid.Size;

            MessageBox.Show(String.Format("Not enough memory for a grid of size ({0}, {1})",
                                          attemptedSize.Width, attemptedSize.Height),
                            Util.GetApplicationName());
        }

        private void SetupCellTypes()
        {
            cellTypes = new Dictionary<string, CellType>();
            foreach(var cellType in cellTypeInitData)
                cellTypes[cellType.name] = cellType;
        }

        public void UpdateStatusLabel(Grid sourceGrid)
        {
            string newStatusStr;

            newStatusStr = String.Format("Map: {0},{1}",
                                         StatusPanelGridSize.Width,
                                         StatusPanelGridSize.Height);

            if(isMouseOverGridPanel && grid.IsInBounds(currCell))
            {
                newStatusStr += String.Format(" | Cursor: {0},{1} | {2}",
                                              currCell.X, currCell.Y,
                                              grid[currCell] == null
                                                ? "null"
                                                : grid[currCell].ToString());
            }

            statusLabel.Text = newStatusStr;
        }

        #region gridPanel Event Handlers
        private void gridPanel_MouseMove(object sender, MouseEventArgs e)
        {
            isMouseOverGridPanel = true;
            grid.GetCellCoords(ref currCell, e.X, e.Y);
            UpdateStatusLabel(grid);
        }

        private void gridPanel_MouseLeave(object sender, EventArgs e)
        {
            isMouseOverGridPanel = false;
            UpdateStatusLabel(grid);
        }
        #endregion

        private void TestButton_Click(object sender, EventArgs e)
        {
            try
            {
                grid[2, 3] = cellTypes["Wall"];
                grid.Refresh();
                statusLabel.Text = "Wall added";
            }
            catch(IndexOutOfRangeException ex)
            {
                statusLabel.Text = "Failed to add Wall: " + ex.Message;
            }
        }

        #region Resize Buttons
        private void DecreaseWidth_Click(object sender, EventArgs e)
        {
            grid.Width -= 1;
        }

        private void IncreaseWidth_Click(object sender, EventArgs e)
        {
            grid.Width += 1;
        }

        private void DecreaseHeight_Click(object sender, EventArgs e)
        {
            grid.Height -= 1;
        }

        private void IncreaseHeight_Click(object sender, EventArgs e)
        {
            grid.Height += 1;
        }
        #endregion

        private void SetSize_Click(object sender, EventArgs e)
        {
            int newWidth  = int.Parse(WidthSet.Text);
            int newHeight = int.Parse(HeightSet.Text);

            if(newWidth <= 0 || newHeight <= 0)
            {
                statusLabel.Text = "Width and Height must be positive.";
                return;
            }

            statusLabel.Text = "Resizing...";
            statusStrip.Refresh();

            // How the fuck do I make this NOT queue up Click events while disabled?!?!?
            // None of these work!
            // Covering it up with another button didn't work either.
            //if(((Button)sender).Enabled == false) return;
            //if(HeightSet.Enabled == false) return;
            //SetSize.Click -= SetSize_Click;
            //SetSize.Hide();
            //button1.Location = SetSize.Location;
            //button1.Show();
            SetSize.Text = "Resizing";
            SetSize.Enabled = false;
            WidthSet.Enabled = false;
            HeightSet.Enabled = false;
            SetSize.Refresh();
            WidthSet.Refresh();
            HeightSet.Refresh();

            grid.Resize(newWidth, newHeight);

            SetSize.Text = "Set Size";
            SetSize.Enabled = true;
            WidthSet.Enabled = true;
            HeightSet.Enabled = true;
            //button1.Location = new Point(button1.Location.X, button1.Location.Y + 30);
            //button1.Hide();
            //SetSize.Show();
            //SetSize.Click += SetSize_Click;
        }

        private void getAutoScroll_Click(object sender, EventArgs e)
        {
            statusLabel.Text = String.Format("scroll position: [{0}, {1}] scroll offset: [{2}, {3}]",
                gridPanel.AutoScrollPosition.X, gridPanel.AutoScrollPosition.Y,
                gridPanel.AutoScrollOffset.X, gridPanel.AutoScrollOffset.Y);
        }

        private void SaveAsButton_Click(object sender, EventArgs e)
        {
            if(saveFileDialog.ShowDialog() == DialogResult.OK)
                grid.Save(saveFileDialog.FileName);
        }

        private void OpenButton_Click(object sender, EventArgs e)
        {
            if(openFileDialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    grid.Open(openFileDialog.FileName);
                }
                catch(GridFileException ex)
                {
                    if(ex is EOFInMapDataException)
                    {
                        MessageBox.Show("The file is incomplete, missing cells filled with nulls.",
                                        Util.GetApplicationName());
                    }
                    else
                    {
                        MessageBox.Show(String.Format("Cannot open file:{0}{1}{0}{0}{2}",
                                                      Environment.NewLine, ex.Uri,
                                                      ex.Message),
                                        Util.GetApplicationName());
                    }
                }
            }
        }
    }
}
