﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using SemiTwist.Utilities;

// Note: The values in ScrollableControl.AutoScrollPosition are negative.
//       Ie, they're the location of the control's CLIENT AREA in relation to its VIEWPORT,
//       and NOT the location of the viewport within the client area.
//
//       Similarly, the ".Location" of a control within a ScrollableControl
//       is also in relation to the VIEWPORT, and NOT the CLIENT AREA.
namespace MazeMaker
{
    public class Grid
    {
        public const int cellWidth = 16;
        public const int cellHeight = 16;
        public const int defaultBorderWidth = 2;

        private CellType[,] cells;
        private CellType defaultCellType;

        private ScrollableControl panel;
        private Panel endMarker;

        // You can disable one or both of these and call Refresh() manually if you plan
        // to make many changes at once and don't want the Grid to waste time redrawing
        // every intermediate change.
        // Note: Turning these off does not guarantee the Grid won't get redrawn
        //       before Refresh() is manually called. Refreshes can still be triggered
        //       by the OS (window unobscured, scroll bars appear/dissapear, etc.)
        public bool autoDrawOnResize = true;
        public bool autoDrawOnSetCell = true;
        
        public event ResizeEventHandler preResize;
        public event ResizeEventHandler postResize;
        public event EventHandler onCellSet;
        public event ExceptionEventHandler onResizeOutOfMemory;

        public delegate void ResizeEventHandler(object sender, Size newSize);
        public delegate void ExceptionEventHandler(object sender, Exception e);

        #region Border
        private Pen borderPen;
        private int _borderWidth;
        private int borderWidth
        {
            get
            {
                return _borderWidth;
            }
            set
            {
                _borderWidth = value;
                borderPen = new Pen(SystemColors.WindowFrame, value);
            }
        }
        #endregion

        #region Accessors
        [System.Runtime.CompilerServices.IndexerName("Cells")]
        public CellType this[int x, int y]
        {
            get
            {
                return cells[x, y];
            }
            set
            {
                cells[x, y] = value;
                if(autoDrawOnSetCell) Draw(x, y);

                if(onCellSet != null)
                    onCellSet(this, new EventArgs());
            }
        }

        [System.Runtime.CompilerServices.IndexerName("Cells")]
        public CellType this[Point p]
        {
            get
            {
                return this[p.X, p.Y];
            }
            set
            {
                this[p.X, p.Y] = value;
            }
        }

        public int Width
        {
            get
            {
                return cells.GetLength(0);
            }
            set
            {
                if(value < 1) value = 1;
                Resize(value, Height);
            }
        }

        public int Height
        {
            get
            {
                return cells.GetLength(1);
            }
            set
            {
                if(value < 1) value = 1;
                Resize(Width, value);
            }
        }

        public Size Size
        {
            get
            {
                return new Size(Width, Height);
            }
            set
            {
                Resize(value.Width, value.Height);
            }
        }
        #endregion

        #region Constructors
        public Grid(ScrollableControl panel)
            : this(panel, null) { }

        public Grid(ScrollableControl panel, CellType defaultCellType)
        {
            this.defaultCellType = defaultCellType;
            this.panel = panel;

            this.panel.AutoScroll = true;
            this.panel.Paint += new PaintEventHandler(panel_Paint);

            borderWidth = defaultBorderWidth;
            InitEndMarker();
            cells = new CellType[0,0];
        }
        #endregion

        // The end marker's purpose is to tell the panel where the grid's bottom-right
        // corner should be, so it can automatically set up it's own scroll bars properly.
        private void InitEndMarker()
        {
            endMarker = new Panel();
            endMarker.Width = 0;
            endMarker.Height = 0;
            endMarker.Location = new Point(borderWidth * 2, borderWidth * 2);
            panel.Controls.Add(endMarker);
        }

        #region IsInBounds
        public bool IsInBounds(Point p)
        {
            return IsInBounds(p.X, p.Y);
        }

        public bool IsInBounds(int x, int y)
        {
            if(x < Width && y < Height && x >= 0 && y >= 0)
                return true;
            else
                return false;
        }
        #endregion

        #region Resize overloads
        public void Resize(int newWidth, int newHeight)
        {
            Resize(newWidth, newHeight, true, true);
        }

        public void Resize(int newWidth, int newHeight, bool preserve)
        {
            Resize(newWidth, newHeight, preserve, true);
        }
        #endregion

        public void Resize(int newWidth, int newHeight, bool preserve, bool fillWithDefault)
        {
            if(preResize != null)
                preResize(this, new Size(newWidth, newHeight));

            CellType[,] newCells;
            try
            {
                newCells = new CellType[newWidth, newHeight];

                if(preserve)
                {
                    // Preserve existing data
                    newCells.SetEach(
                        new int[] { 0, 0 },
                        new int[] {Math.Min(newWidth, this.Width),
                               Math.Min(newHeight, this.Height)},
                        GetValueAtBuiltIn.Copy, cells);

                    if(fillWithDefault)
                    {
                        // Set new cells to default
                        if(newWidth > this.Width) // Width Increasing
                        {
                            newCells.SetEach(
                                new int[] { this.Width, 0 },
                                new int[] { newWidth, newHeight },
                                GetValueAtBuiltIn.Fill, defaultCellType);
                        }

                        if(newHeight > this.Height) // Height Increasing
                        {
                            newCells.SetEach(
                                new int[] { 0, this.Height },
                                new int[] { Math.Min(newWidth, this.Width), newHeight },
                                GetValueAtBuiltIn.Fill, defaultCellType);
                        }
                    }
                }
                else if(fillWithDefault)
                {
                    newCells.SetEach(GetValueAtBuiltIn.Fill, defaultCellType);
                }
            }
            catch(OutOfMemoryException e)
            {
                if(onResizeOutOfMemory != null)
                {
                    onResizeOutOfMemory(this, e);
                    return;
                }
                else
                    throw e;
            }

            // Commit and Refresh
            cells = newCells;
            endMarker.Location = new Point(
                (newWidth  * cellWidth)  + (borderWidth * 2) + panel.AutoScrollPosition.X,
                (newHeight * cellHeight) + (borderWidth * 2) + panel.AutoScrollPosition.Y);

            if(autoDrawOnResize) Refresh();
            if(postResize != null)
                postResize(this, new Size(newWidth, newHeight));
        }

        // Just for convenience
        public void Refresh()
        {
            panel.Refresh();
        }

        private void panel_Paint(object sender, PaintEventArgs e)
        {
            Draw();
        }

        private void Draw()
        {
            Graphics g = panel.CreateGraphics();

            g.DrawRectangle(borderPen,
                            (borderWidth / 2) + panel.AutoScrollPosition.X,
                            (borderWidth / 2) + panel.AutoScrollPosition.Y,
                            (endMarker.Location.X - borderWidth) - panel.AutoScrollPosition.X,
                            (endMarker.Location.Y - borderWidth) - panel.AutoScrollPosition.Y);

            int startX = GetCellCoordX(0) - 1;
            int startY = GetCellCoordY(0) - 1;
            int endX = GetCellCoordX(panel.Size.Width - borderWidth) + 1;
            int endY = GetCellCoordY(panel.Size.Height - borderWidth) + 1;

            if(startX < 0) startX = 0;
            if(startY < 0) startY = 0;
            if(endX > Width) endX = Width;
            if(endY > Height) endY = Height;

            // This is slightly faster than using DoEach. The extra speed is important for scrolling.
            for(int y = startY; y < endY; y++)
            for(int x = startX; x < endX; x++)
            {
                if(cells[x, y] != null)
                {
                    // Note: the values in AutoScrollPosition are negative
                    g.DrawImage(cells[x, y].bitmap,
                                borderWidth + (x * cellWidth) + panel.AutoScrollPosition.X,
                                borderWidth + (y * cellHeight) + panel.AutoScrollPosition.Y);
                }
            }
        }

        private void Draw(int x, int y)
        {
            Graphics g = panel.CreateGraphics();

            if(cells[x, y] != null)
            {
                // Note: the values in AutoScrollPosition are negative
                g.DrawImage(cells[x, y].bitmap,
                            (x * cellWidth) + panel.AutoScrollPosition.X,
                            (y * cellHeight) + panel.AutoScrollPosition.Y);
            }
        }

        #region File IO
        public const string fileTypeID = "MMGF";
        public const ushort fileTypeVersion = 0x0100;
        public void Save(string filename)
        {
            using(var file =
                  new BinaryWriter(new FileStream(filename, FileMode.Create),
                                   new UnicodeEncoding(false, false, false))
                 )
            {
                foreach(var c in fileTypeID)
                    file.Write((byte)c);
                file.Write(fileTypeVersion);
                file.Write(Width);
                file.Write(Height);

                //TODO: Fix lambda version (ie, add versions of DoAt/GetValueAt without extraRef)
                cells.DoEach(
                delegate(int[] i, Array a, ref object r, object[] e)
                //(i, a, e) =>
                {
                    file.Write(((CellType)(a.GetValue(i))).ID);
                });
            }
        }

        //TODO: Make unittests
        public void Open(string filename)
        {
            byte[] readFileTypeIDData;
            char[] readFileTypeID = new char[4];
            ushort readFileTypeVersion;
            int width;
            int height;

            using(var file =
                  new BinaryReader(new FileStream(filename, FileMode.Open),
                                   new UnicodeEncoding(false, false, false))
                 )
            {
                try
                {
                    readFileTypeIDData = file.ReadBytes(fileTypeID.Length);
                    if(readFileTypeIDData.Length < fileTypeID.Length)
                        throw new EndOfStreamException();

                    readFileTypeID = new char[readFileTypeIDData.Length];
                    readFileTypeID.SetEach(GetValueAtBuiltIn.Copy, readFileTypeIDData);
                    if(new string(readFileTypeID) != fileTypeID)
                        throw new TypeIDException(filename);

                    readFileTypeVersion = file.ReadUInt16();
                    if(readFileTypeVersion > fileTypeVersion)
                        throw new FutureVersionException(filename);
                    if(readFileTypeVersion != fileTypeVersion)
                        throw new UnknownVersionException(filename);

                    width = file.ReadInt32();
                    height = file.ReadInt32();
                    if(width <= 0 || height <= 0)
                        throw new UnsupportedDimentionsException(filename);

                    Resize(width, height, false, false);

                    //TODO: Fix lambda version (ie, add versions of DoAt/GetValueAt without extraRef)
                    try
                    {
                        cells.SetEach(
                        delegate(int[] i, Array a, ref object r, object[] e)
                        //(i, a, e) =>
                        {
                            //TODO: Tidier way to handle null condition
                            if(onCellSet != null)
                                onCellSet(this, null);

                            return CellType.FromID(file.ReadByte());
                        });
                    }
                    catch(EndOfStreamException)
                    {
                        throw new EOFInMapDataException(filename);
                    }
                    finally
                    {
                        Refresh();
                    }
                }
                catch(EndOfStreamException)
                {
                    throw new EOFInHeaderException(filename);
                }
            }
        }
        #endregion

        #region Coordinate Conversion
        // Functions to convert from pixel coordinates to cell coordinates
        // Note: Truncation is NOT equivilent to Flooring when negative values are involved.

        public int GetCellCoordX(int x)
        {
            return (int)Math.Floor((double)((x - borderWidth) - panel.AutoScrollPosition.X) /
                                   (double)Grid.cellWidth);
        }

        public int GetCellCoordY(int y)
        {
            return (int)Math.Floor((double)((y - borderWidth) - panel.AutoScrollPosition.Y) /
                                   (double)Grid.cellHeight);
        }

        //public static Point GetCellCoords(int x, int y)
        //{
        //    return new Point(GetCellCoordX(x), GetCellCoordY(y));
        //}

        //public static Point GetCellCoords(Point coord)
        //{
        //    return GetCellCoords(coord.X, coord.Y);
        //}

        public void GetCellCoords(ref Point cellCoords, int pixelX, int pixelY)
        {
            cellCoords.X = GetCellCoordX(pixelX);
            cellCoords.Y = GetCellCoordY(pixelY);
        }

        public void ConvertToCellCoords(ref Point coord)
        {
            coord.X = GetCellCoordX(coord.X);
            coord.Y = GetCellCoordY(coord.Y);
        }
        #endregion
    }
}
